#include <stdio.h>
#include <ctype.h>
#include "toptext.h"

#define ESC 0x1b
#define ishigh(c) ((c) >= 0x80)

typedef enum { BIN, ASCII, JIS, EUC, SJIS } Code;

// $BJ8;z%3!<%I$rH=Dj(B
static Code checkcode(unsigned char *s)
{
  Code code = ASCII;
  enum {LOW, HIGH} lh = LOW;
  unsigned char c;

  for(;*s;s++){
    c = *s;
    if(lh == LOW){
      if(c == ESC){ code = JIS; }
      else if(isspace(c) || isprint(c)){ }
      else if(c >= 0x80){
        if(code == JIS) return BIN;
        lh = HIGH;
      }
      else { return BIN; }
    }
    else { // $B%3!<%I$K$h$j(BSJIS/EUC$BH=Dj(B
      if(c >= 0x20 && c <= 0xa0){ code = SJIS; }
      else if(c >= 0xa1 && c <= 0xfe){
        if(code != SJIS) code = EUC;
      }
      else { return BIN; }
      lh = LOW;
    }
  }
  return code;
}

typedef enum { SSTART, SESC, SESC2, SHIGH } State;

// $B@hF,$+$i(B1000$B%P%$%H$r(BEUC$B$G<hF@(B
unsigned char *toptext(unsigned char *s)
{
  State state = SSTART;
  static unsigned char buf[2000];
  unsigned char *p = buf;
  unsigned char c,c1,c2;
  int jis = 0;
  Code code;

  code = checkcode(s);
  if(code == BIN) return "";

  for(;p-buf<1000 && *s;s++){
    c = *s;
    if(!jis && isupper(c)) c = tolower(c);
    switch(state){
    case SSTART:
      if(c == ESC){ state = SESC; }
      else if(isspace(c)){ }
      else if(c >= 0x80){
        c1 = c;
        state = SHIGH;
      }
      else { // JIS->EUC$BJQ49(B
        if(jis) *p++ = (c | 0x80);
        else    *p++ = c;
      }
      break;
    case SESC:
      if(c == '$'){ jis = 1; state = SESC2; }
      else if(c == '('){ jis = 0; state = SESC2; }
      else { state = SSTART; }
      break;
    case SESC2:
      state = SSTART;
      break;
    case SHIGH:
      c2 = c;
      if(code == SJIS){ // SJIS->EUC$BJQ49(B
        if(c1 >= 0xe0) c1 -= 0x40;
        if(c2 >= 0x80) c2--;
        c1 = (c1-0x81) * 2 + (c2>=0x9e ? 1 : 0) + 0xa1;
        c2 = (c2 >= 0x9e ? c2-0x9e : c2-0x40) + 0xa1;
      }
      *p++ = c1;
      *p++ = c2;
      state = SSTART;
      break;
    default:
      break;
    }
  }
  *p = '\0';
  return buf;
}
